/*
	Copyright (c) 2001 Maurice Gittens <maurce@gittens.nl>

   This software is distributed under the GNU General
   Public License.
	See the file COPYING for more information on usage and
	redistribution of this file and for a DISCLAIMER OF ALL
	WARRANTIES.

	Written by Maurice Gittens <maurice@gittens.nl>.
*/

#include <tiny_codegenerator.h>
#include <tiny_parserGenerator.h>
#include <tiny_lexer.h>
#include <tiny_action.h>
#include <tiny_genericsymbol.h>
#include <unistd.h>
#include <fstream>
#include <config.h>
#include <algorithm>
#include <cctype>

using namespace std;

//
//
//
void tiny_codeGenerator::genHeaderTokens(const char *s)
{
   header << s;
}

//
//
//
void tiny_codeGenerator::genClassName()
{
   header << "\n\nclass " << scanner.getGrammarName() << " ";
}

//
//
//
void tiny_codeGenerator::generateCode()
{
   genCodeHeader();
   genIncludeHeader();

   genConstants();
   genNamesOfSymbols();

   genCSRuleTable();
   genCSRuleOffsetTable();
   genNonTerminalOffsetTable();
   genGenericActionOffsetTable();

   genGenerics();

   genIncludeFooter();
	genCodeFooter();

   genActions();
}

//
//
//
void tiny_codeGenerator::traverse(void (tiny_codeGenerator::*f)(const tiny_action &))
{
	int count = scanner.getActionCount();
	for(int i = 0; i < count; i++)
		(this->* f)(scanner.getAction(i));
}

//
//
//
void tiny_codeGenerator::traverse(void (tiny_codeGenerator::*f)(tiny_rewritableSymbol &))
{
	int count = scanner.getNonTerminalCount();
	for(int i = 0; i < count; i++)
		(this->* f)(scanner.getNonTerminal(i));
}

//
//
//
void tiny_codeGenerator::traverse(void (tiny_codeGenerator::*f)(const tiny_rewritableSymbol &))
{
	int count = scanner.getNonTerminalCount();
	for(int i = 0; i < count; i++)
		(this->* f)(scanner.getNonTerminal(i));
}

//
//
//
void tiny_codeGenerator::traverse(void (tiny_codeGenerator::*f)(const tiny_grammarSymbol &))
{
	int count = scanner.getTerminalCount();
	for(int i = 0; i < count; i++)
		(this->* f)(scanner.getTerminal(i));
}

void tiny_codeGenerator::genCodeHeader()
{
	static const char* text1 =
	"/*\n"
	" This file was generated by the Tiny parser generator version " VERSION "\n"
	" from the input file: ";

	static const char* text2 = " All output generated by the this version of Tiny must be distributed\n"
	          " under the GNU General Public Licence version 2 or higher.\n\n*/\n";
	
   header << text1 << pg.getInputFileName() << ".\n\n" << text2 << endl;
   source << text1 << pg.getInputFileName() << ".\n\n" << text2 << endl;
	source << "#include <" << pg.getHeaderFileName() << ">\n\n" << endl;
}

void tiny_codeGenerator::genCodeFooter()
{
   source << scanner.getSourceCodeBlock();
}

//
//
//
void tiny_codeGenerator::genIncludeHeader()
{
   std::string gn = scanner.getGrammarName();

//   std::transform(gn.begin(), gn.end(),
//                  gn.begin(), std::toupper);

	header << "#ifndef " << gn << "_SEEN\n"
	       << "#define " << gn << "_SEEN\n\n";
	
	static const char text[] =
	"/*\n"
	"\tThis file was generated by the Tiny parser generator version " VERSION "\n"
	"\tfrom the input file: ";
	
	header << text << pg.getInputFileName() << ".\n" << endl;
	
	header << "\tAll output generated by the this version of Tiny must be distributed\n"
	          "\tunder the GNU General Public Licence version 2 or higher\n\n*/\n";

	header << "\n#include <tiny_header.h> \n\n";

	string ih = scanner.getInheritBlock();
	
	if (ih.empty())
	   ih = string("TinyParser<" + scanner.getTypeName() + ">");

	header << "class " << scanner.getGrammarName() << " : "
          << ih
	       << "\n{\n"
	       << "public:\n" <<
          "\tconst char** getTerminalNames() const { return terminalNames; }\n"
          "\tconst char** getNonTerminalNames() const { return symbolNames; }\n"
          "\tconst unsigned short* getRuleTable() const { return ruleTable; }\n"
          "\tconst unsigned short* getCsRuleOffsetTable() const { return csRuleOffsetTable; }\n"
          "\tconst unsigned short* getSymbolOffsetTable() const { return symbolOffsetTable; }\n"
          "\tconst unsigned short* getGenericsTable() const { return genericsTable; }\n"
          "\tconst unsigned short getStartSymbol() const { return startSymbol; }\n"
          "\tconst unsigned short getErrorSymbol() const { return errorSymbol; }\n"
          "\tconst unsigned short getPlaybackSymbol() const { return playbackSymbol; }\n"
          "\tconst unsigned short getAnySymbol() const { return anySymbol; }\n"
          "\tunsigned getNonTerminalCount() const { return nonTerminalCount; }\n"
          "\tunsigned getTerminalCount() const { return terminalCount; }\n\n"
          "\tint action(unsigned);\n\n"
          "\tvoid getToken(Symbol<" << scanner.getTypeName() <<
          ">&) const;\n"
          "private:\n"
          "\tstatic const char* terminalNames[];\n"
          "\tstatic const char* symbolNames[];\n"
          "\tstatic const unsigned short ruleTable[];\n"
          "\tstatic const unsigned short csRuleOffsetTable[];\n"
          "\tstatic const unsigned short symbolOffsetTable[];\n"
          "\tstatic const unsigned short startSymbol;\n"
          "\tstatic const unsigned short anySymbol;\n"
          "\tstatic const unsigned short playbackSymbol;\n"
          "\tstatic const unsigned short errorSymbol;\n"
          "\tstatic const unsigned short nonTerminalCount;\n"
          "\tstatic const unsigned short terminalCount;\n\n"
          "\tstatic const unsigned short genericsTable[];\n\n";

   header << scanner.getHeaderCodeBlock();
}

void tiny_codeGenerator::genIncludeFooter()
{
	header << "\n};\n\n#endif\n";
}

//
//
//
void tiny_codeGenerator::genNamesOfSymbols()
{
	header << "\n\tenum Symbols\n\t{\n";

	if (scanner.getLocalTerminalCount() || scanner.getGlobalTerminalCount())
	{
		genCharTemplatePrefix();
		source << "terminalNames[]={\n";
		traverse(&tiny_codeGenerator::genTerminalName);
		source << "\n};\n";
	}

	if (scanner.getNonTerminalCount())
	{
		genCharTemplatePrefix();
		source << "symbolNames[]=\n\t{\n";
		traverse(&tiny_codeGenerator::genNonTerminalName);
		source << "\n};\n";
	}

	header << "\t};\n\n";
	header << "private:";

}

//
//
//
void tiny_codeGenerator::genTerminalName(const tiny_grammarSymbol &t)
{
	source << "\t\"" << t.getName() << "\", /* = " << t.getEncodedId() << " */\n";
	header << "\t\t" << t.getName() << " = " << t.getEncodedId() << ",\n";
}

//
//
//
void tiny_codeGenerator::genNonTerminalName(const tiny_rewritableSymbol& s)
{
	source << "\t\"" << s.getName()
	       << "\", /* = " << s.getEncodedId() << " */\n";
	header << "\t\t" << s.getName()
	       << " = " << s.getEncodedId();

	// this because IBM C/C++ 2.1 doest handle enums ending with a comma correctly
	if (s.getId() < (scanner.getNonTerminalCount() - 1))
		header << ",\n";
	else
		header << "\n";

}

//
//
//
void tiny_codeGenerator::genConstants()
{
	source << "\n\n";
	genTemplatePrefix();
	source << "startSymbol = " <<
	scanner.getStartSymbol() << ";\n";

	genTemplatePrefix();
	source << "errorSymbol = " <<
	scanner.getErrorSymbol() << ";\n";

	genTemplatePrefix();
	source << "playbackSymbol = " <<
	scanner.getPlaybackSymbol() << ";\n";

	genTemplatePrefix();
	source << "anySymbol = " <<
	scanner.getAnySymbol() << ";\n";

	genTemplatePrefix();
	source << "terminalCount = " <<
	scanner.getTerminalCount() << ";\n";

	genTemplatePrefix();
	source << "nonTerminalCount = " <<
	scanner.getNonTerminalCount() << ";\n";

	source << "\n\n";
}

//
//
//
void tiny_codeGenerator::genGenericActionOffsetTable()
{
	genTemplatePrefix();
	source << "genericsTable[]={\n\t";
	traverse(&tiny_codeGenerator::genGenericActionOffset);
	source << "\n};\n";
}

//
//
//
void tiny_codeGenerator::genNonTerminalOffsetTable()
{
	genTemplatePrefix();
	source << "symbolOffsetTable[]={\n\t";
	traverse(&tiny_codeGenerator::genNonTerminalOffset);
	source << "\n};\n";
}

//
//
//
void tiny_codeGenerator::genGenericActionOffset(const tiny_rewritableSymbol&  s)
{
   if (const tiny_genericSymbol* sym =
       dynamic_cast<const tiny_genericSymbol*>(&s))
   {
      assert(sym->getType() == tiny_grammarSymbol::generic);
	   source << makeId(sym->getAction().getId()) << ", ";
   }
   else
   {
	   source << (unsigned short)-1 << ", ";
   }
}

//
//
//
void tiny_codeGenerator::genNonTerminalOffset(const tiny_rewritableSymbol&  s)
{
   source << s.getOffset() << ", ";

   if (!(s.getId() % 10) && s.getId())
      source << endl;
}

void tiny_codeGenerator::genGenerics()
{
	traverse(&tiny_codeGenerator::genGenerics);
}

void tiny_codeGenerator::genGenerics(const tiny_rewritableSymbol& nt)
{
	if (nt.getType() != tiny_grammarSymbol::generic)
      return;

   header << "\nbool " << scanner.getGrammarName() << "::"
          << nt.getName() << "_nt_();\n";

   source << "\nbool " << scanner.getGrammarName() << "::"
          << nt.getName() << "_nt_()\n{\n";

	source << "if (!rewriteSymbol(" << nt.getEncodedId() << "))\n";
   const char* snippet1 = ""
	     "{\n"
	     "   if (token(-1) == getErrorSymbol())\n"
		        "throw tiny_exception(\"syntax error in error handler\");\n"

	        "pushL(current);"
	        "current.id = getErrorSymbol();"
	        "pushR(current);"
	     "}";
	
   source << snippet1 << endl;
   source << "}\n";
}

//
//
//
void tiny_codeGenerator::genCSRuleTable()
{
   offset = 0;

   genTemplatePrefix();
   source << "ruleTable[]={" << endl;
   traverse(&tiny_codeGenerator::genCSRule);
   source << "\n};\n";
}

//
//
//
void tiny_codeGenerator::genCSRule(tiny_rewritableSymbol& s)
{
//	if (s.getType() == tiny_grammarSymbol::generic)
//      return;
//   else
	{
      const int ruleCount = s.getRuleCount();
	   for(int i = 0; i < ruleCount; i++)
      {
	      tiny_principle &rule = s.getRule(i);
		   tiny_symbolString &lhs = s.getLHS(i);
   	   const int lhsLength = lhs.length();

	      int dotIndex = lhs.getDotIndex();
		   int counter = dotIndex;

         rule.setOffset(offset);

	      source << endl << '\t';

		   // emit lhs pre context
   		while(counter--)
	   	{
		   	source << lhs.getSymbol(counter) << ", ";
			   offset++;
   		}

	   	source << 0 << ", ";
		   offset++;

   		// emit lhs post context; dotIndex + 2 because we don't need to match the current symbols
	   	for(counter = dotIndex + 2; counter < lhsLength; counter++)
		   {
   			source << lhs.getSymbol(counter) << ", ";
	   		offset++;
		   }

   		source << 0 << ", ";
	   	offset++;

		   for(int j = 0 ; j < rule.getAlternativeCount(); ++j)
   		{
	   		tiny_principle::RightHandSide &rhs = rule.getRHS(j);
		   	tiny_symbolString &symStr = rhs.getString();

			   // Add selection set code here
   			if (!rhs.hasSelectionAction())
	   		{
		   		if ( rhs.isNullable() ||	 		// nullable rule
			   		 rhs.getFirstSet()[makeId(scanner.getAnySymbol())] ||   	// any in first set
   					 rhs.getFirstSet().empty() )
	   			{
		   			source << ALWAYS_REWRITE << ", ";
			   		offset++;
				   } else 					// rhs has a first set
   				{
	   				source << SELECTIONSET_MASK << ", ";
		   			offset++;

			   		vector<bool>& bitset = rhs.getFirstSet();
				   	int cardinality;
					   for(unsigned k = 0; k < bitset.size(); ++k)
   					{
	   					if(bitset[k])
		   				{
			   				source << makeTerminal(k) << ", ";
				   			offset++;
					   		++cardinality;
						   }

   						if (!(cardinality % 20))
	   						source << "\n\t";

			   		}

				   	// terminate the selection set with a 0 character
					   source << 0 << ", ";
   					offset++;
	   			}
		   	}

			   emitSymbolString(symStr);
			
   			if (!rhs.getStringToRecord().empty()) // a string to record?
	   		{
   	   		emitSymbolString(rhs.getStringToRecord());
			   }
			
   			source << (unsigned short)-1 << ", ";
	   		offset++;
		   }
   	}

	   source << (unsigned short)-1 << ", ";
   	offset++;
   }
}

//
//
//
void tiny_codeGenerator::genCSRuleOffsetTable()
{
	offset = 0;

	genTemplatePrefix();
	source << "csRuleOffsetTable[]={\n";
	traverse(&tiny_codeGenerator::genCSRuleOffset);
	source << "\n};\n";
}

//
//
//
void tiny_codeGenerator::genCSRuleOffset(tiny_rewritableSymbol& s)
{
	s.setOffset(offset);
	source  << "\t";

	for(unsigned int i = 0; i < s.getRuleCount(); i++)
	{
		const tiny_principle &rule = s.getRule(i);

		source << rule.getOffset() << ", ";
		offset++;

		if (i && !(i % 10))
			source << "\n\t";
	}

	source << (unsigned short)-1 << ", \n";
	offset++;
}

//
//
//
void tiny_codeGenerator::genActions()
{
	source << "\n\n";

	source << "int " << scanner.getGrammarName();
	source << "::action(unsigned __a__)\n{\nswitch(makeId(__a__)) {\n";

	if (scanner.getActionCount())
	{
		traverse(&tiny_codeGenerator::genAction);
	}
	else
	{
		source << ";\n";
	}

	source << "}\nreturn 1;\n";
	source << "}\n\n";
}

//
//
//
void tiny_codeGenerator::genAction(const tiny_action& a)
{
	source << "\tcase " << a.getId() << ": {\n";

	genLineInfo(a.getLineNo(),0);  // 0 is source file

	source << a.getString() << '\n';

	source << "\t}\n\tbreak;\n";
}

//
//
//
void tiny_codeGenerator::genLineInfo(int line,int to)
{
	if (pg.getLineNumbers())
	{
		if (to)
			header << "#line " << line << " \"" << pg.getInputFileName() << "\"\n\t\t";
		else
			source << "#line " << line << " \"" << pg.getInputFileName() << "\"\n\t\t";
	}
}

//
//
//
void tiny_codeGenerator::genTemplatePrefix()
{
	source << "const unsigned short " << scanner.getGrammarName() << "::";
}

//
//
//
void tiny_codeGenerator::genCharTemplatePrefix()
{
	source << "const char *" << scanner.getGrammarName() << "::";
}

void tiny_codeGenerator::emitSymbolString(const tiny_symbolString &ss)
{
   int length = ss.length();
   int dotIndex = ss.getDotIndex();
   int counter = dotIndex;

	// emit rhs pre context
	int k;
	for(k = 0; k < counter; ++k)
	{
		source << ss.getSymbol(k) << ", ";
		++offset;
	}

	source << 0 << ", ";
	++offset;

	// emit rhs post context
	for(k = length - 1; k > counter; --k)
	{
		source << ss.getSymbol(k) << ", ";
		++offset;
	}

	source << 0 << ", ";
	++offset;
}

