#ifndef TINY_HEADER_SEEN
#define TINY_HEADER_SEEN

#include <vector>
#include <tiny_stack.h>
#include <tiny_util.h>
#include <iostream>
#include <stdexcept>
#include <cstdio>

template <class T>
struct Symbol
{
     Symbol() {}
     Symbol(short i, T& a) { id = i; attr = a; }
     short id;
     T attr;
};

template<typename AttributeType>
class TinyParser
{
public:

protected:

     virtual const char** getTerminalNames() const = 0;
     virtual const char** getNonTerminalNames() const = 0;
     virtual unsigned getTerminalCount()  const = 0;
     virtual unsigned getNonTerminalCount() const = 0;

     virtual const unsigned short* getRuleTable() const = 0;
     virtual const unsigned short* getCsRuleOffsetTable() const = 0;
     virtual const unsigned short* getSymbolOffsetTable() const = 0;
     virtual const unsigned short* getGenericsTable() const = 0;

     virtual const unsigned short getStartSymbol() const = 0;
     virtual const unsigned short getAnySymbol() const = 0;
     virtual const unsigned short getPlaybackSymbol() const = 0;
     virtual const unsigned short getErrorSymbol() const = 0;

     int getLookAhead()
     {
   	  if (consume)
	     {
	        getToken(lookAhead);
	        consume = 0;
	     }
        return lookAhead.id;
     }

     AttributeType& lookAheadAttr()
     {
	     if (consume)
	     {
	        getToken(lookAhead);
	        consume = 0;
	     }

	     return lookAhead.attr;
     }

     const AttributeType& currentAttr() const { return current.attr; }
     AttributeType& currentAttr()   { return current.attr; }
     const unsigned short& currentToken()   { return current.id; }

     Symbol<AttributeType> lookAhead;
     Symbol<AttributeType> current;

     int rewriteSymbol(int);
     int rewriteGenericSymbol(int);

     int match(unsigned short t)
     {
	     return (lookAhead() == t) ? (current = lookAhead, consume = 1) : 0;
     }

     void advance()
     {
 	     lookAhead();
	     current = lookAhead;
        consume = 1;
     }

     unsigned short token(int i)  { return i < 0 ? tape->accessL(-i - 1).id : tape->accessR(i- 1).id;}
     AttributeType& attr(int i) { return i < 0 ? tape->accessL(-i - 1).attr: tape->accessR(i - 1).attr; }

     const AttributeType& lhsAttr(int i) const { return i < 0 ?preContext[-i-1].attr : postContext[i-1].attr; }
     const short lhsToken(int i) const { return i < 0 ? preContext[-i-1].id :postContext[i-1].id; }

     int itemCountL() { return tape->itemCountL(); }
     int itemCountR() { return tape->itemCountR(); }
     int itemCount() { return tape->itemCount(); }

     bool emptyR() const { return tape->emptyR(); }

     void popR(int i) { tape->popR(i); }
     void popL(int i) { tape->popL(i); }

     void popR(Symbol<AttributeType> &s) { s = tape->popR();}
     void popL(Symbol<AttributeType> &s) { s = tape->popL(); }

     void pushR(const Symbol<AttributeType> &s) { tape->pushR(s); }
     void pushL(const Symbol<AttributeType> &s) { tape->pushL(s); }

     short peekR(int i) const { return tape->peekR(i).id; }
     short peekL(int i) const { return tape->peekL(i).id; }

     const Symbol<AttributeType>& accessR(int i) const { return tape->accessR(i);}
     const Symbol<AttributeType&>& accessL(int i) const { return tape->accessL(i);}
     const Symbol<AttributeType&>& access(int i) const { return i < 0 ? tape->accessL(-i- 1) : (!i ? current : tape->accessR(i - 1)); }

     Symbol<AttributeType>& accessR(int i)    { return tape->accessR(i);}
     Symbol<AttributeType&>& accessL(int i)    { return tape->accessL(i);}
     Symbol<AttributeType&>& access(int i) { return i < 0 ? tape->accessL(-i- 1) : (!i ? current : tape->accessR(i - 1)); }

     void redo() { tape->pushR(current); }

     virtual void getToken( Symbol<AttributeType> &) const = 0;
     virtual void putToken( const Symbol<AttributeType> &) {  } // Hmm, this probably not needed

private:

     char nameBuf[5];

     short trace__;
     short traceTerminals__;
     short verbose__;
     std::vector<bool> bits;

     std::vector<Symbol<AttributeType> > preContext;
     std::vector<Symbol<AttributeType> >postContext;

     bool consume;

     const unsigned short *tablePointer;
     const unsigned short *rulePointer;
     int  preContextLength;
     int  postContextLength;

     void saveLHS();

     void rewiteNonTerminal(int);
     void rewrite();
     int matchLHS()
     {
	     return (matchPreContext() && matchPostContext()) ? RHSSelected() : 0;
     }

     int matchPreContext();
     int matchPostContext();
     int RHSSelected();

     virtual int action(unsigned i) = 0;

public:
     TinyParser(unsigned int mrhs = 32, unsigned int tapesize = 1024)
     : tape(tapesize, mrhs), preContext(mrhs), postContext(mrhs)
     {
         verbose__ = trace__ = traceTerminals__ = 0;
     }
     virtual int step();
     virtual void begin(const AttributeType& in, int ss)
     {
	     consume = 1;

	     current.id = (short)ss;
	     current.attr = in;

	     pushR(current);
     }

     virtual void end() {}
     virtual void run() { AttributeType dummy; run(getStartSymbol(),dummy); }
     virtual void run(const AttributeType& in) { run(getStartSymbol(),in); }
     virtual void run(int , const AttributeType&);

     void verbose(bool on = true) { verbose__ = on; }

     void trace(bool on = true) { trace__ = on; }
     void traceAll(bool on = true)
     {
        if (on)
        {
           traceTerminals();
           // setall bits to true bits.Set();
        }
        else
        {
           traceTerminals(false);
           // clear all bits in bits.Clear
        }
     }

     void trace(int s)   { bits.Set(makeId(s)); }
     void noTrace(int s)      { bits.Clear(makeId(s)); }

     bool getTraceTerminalsFlag() const { return traceTerminals__;   }
     void traceTerminals(bool on = true) { traceTerminals__ = on;   }
     int isTracing() const { return trace__; }
     bool traceSymbol(int s) { return bits[makeId(s)]; }
     bool getVerbose() const { return verbose__; }
     virtual void debugString(const char *s) { std::cout << s;  }
     virtual void debugInt(int i) { std::cout << i; }

     const char *getSymbolName(unsigned i);

     virtual void Init() { reset(); }
     virtual void reset()	{ consume = 1; trace__ = verbose__ = false; traceTerminals__ = 0; tape->reset(); }

     TINY_ContextMachine< Symbol<AttributeType> >  tape;
};

template<typename AttributeType>
int TinyParser<AttributeType>::rewriteSymbol(int s)
{
   rulePointer = &getCsRuleOffsetTable()[getSymbolOffsetTable()[makeId(s)]];

   for(; *rulePointer != (unsigned short)-1; rulePointer++)
   {
	   if (matchLHS())
	   {
	      saveLHS();

         if (isTracing() && traceSymbol(current.id))
         {
            int i;
            std::vector<Symbol<AttributeType> >::const_iterator p1 = preContext.begin();
            std::vector<Symbol<AttributeType> >::const_iterator p2 = postContext.begin();

            p1 += (i = preContextLength);

            while(i--)
            {
               debugString(getSymbolName((*--p1).id));
               debugString(" ");
            }

            debugString(". ");
            for(i = 0; i < postContextLength + 1; i++)
            {
               debugString(getSymbolName((*p2++).id));
               debugString(" ");
            }
         }

	      rewrite();
	      return 1;
	   }
   }

   return 0;
}

template<typename AttributeType>
int TinyParser<AttributeType>::rewriteGenericSymbol(int s)
{
   rulePointer = &getCsRuleOffsetTable()[getSymbolOffsetTable()[makeId(s)]];

   for(; *rulePointer != (unsigned short)-1; rulePointer++)
   {
      if (matchPreContext() && matchPostContext())
	   {
         saveLHS();

         /* find a applicable RHS by searching them in order */

         if (isTracing() && traceSymbol(current.id))
         {
            int i;
            std::vector<Symbol<AttributeType> >::const_iterator p1 = preContext.begin();
            std::vector<Symbol<AttributeType> >::const_iterator p2 = postContext.begin();

            p1 += (i = preContextLength);

            while(i--)
            {
               debugString(getSymbolName((*--p1).id));
               debugString(" ");
            }

            debugString(". ");
            for(i = 0; i < postContextLength + 1; i++)
            {
               debugString(getSymbolName((*p2++).id));
               debugString(" ");
            }
         }

	      rewrite();
	      return 1;
	   }
   }

   return 0;
}

template<typename AttributeType>
int TinyParser<AttributeType>::matchPreContext()
{
     tablePointer = &getRuleTable()[*rulePointer];

     for (preContextLength = 0; *tablePointer; preContextLength++, tablePointer++)
	  if (*tablePointer != peekL(preContextLength) &&
	       !isVariable(*tablePointer))
	       return 0;

     tablePointer++;
     return 1;
}

template<typename AttributeType>
int TinyParser<AttributeType>::matchPostContext()
{
     int rv = 1;
     for (postContextLength = 0; *tablePointer; postContextLength++, tablePointer++)
     {
	     if (*tablePointer != peekR(postContextLength) &&
	         !isVariable(*tablePointer))
	     {
	        if (isAction(*tablePointer))
           {
              pushR(current);
		        rv = action(*tablePointer);
              postContextLength--;
		        popR(1);
           }
	        else
		        return 0;
        }
     }

     tablePointer++;
     return rv;
}

template<typename AttributeType>
int TinyParser<AttributeType>::RHSSelected()
{
   while(*tablePointer != (unsigned short)-1)
   {
	   if (isAction(*tablePointer))
	   {
	      pushR(current);
	      int result = action(*tablePointer++);
	      popR(1);

	      if (result)
			   return 1;
	      else
		      while(*tablePointer++ != (unsigned short)-1)
            ;
	   }
	   else if (isAlwaysRewrite(*tablePointer))
	   {
		   tablePointer++;
		   return 1;
	   } else if (isSelectionSet(*tablePointer))
      {
		   int la = getLookAhead();
		   tablePointer++;

		   while(*tablePointer && la != *tablePointer)
			   tablePointer++;

		   if (*tablePointer)
		   {
            while(*tablePointer++)
				   ;
			   return 1; // select this rhs
		   }
		   else
            while(*tablePointer++ != (unsigned short)-1)
			      ;

      } else
	   {
         assert( 0 == "this shouldn't happen");
	   }
   }

   return 0;
}

template<typename AttributeType>
void TinyParser<AttributeType>::saveLHS()
{
     int i;
     for(i = 0; i < preContextLength ; i++)
	     preContext[i] = tape->accessL(i);

     postContext[0] = current;

     for(i = 1; i <= postContextLength ; i++)
	     postContext[i] = tape->accessR(i - 1 );
}

template<typename AttributeType>
const char *TinyParser<AttributeType>::getSymbolName(unsigned i)
{
   const char** terminalNames = getTerminalNames();

   if (!i)
      return ". ";

   const int id = makeId(i);

   switch(symbolType(i))
   {
   case TINY_TERMINAL_MASK :
      return getTerminalNames()[id];
   case TINY_NONTERMINAL_MASK :
   case TINY_GENERIC_MASK :
         return getNonTerminalNames()[id];
   case TINY_ACTION_MASK :
         sprintf(nameBuf,"#%d", id);
   return nameBuf;
   default:
      return "\"bad token id\"";
   }
}

template<typename AttributeType>
void TinyParser<AttributeType>::
run(int ss, const AttributeType& in)
{
     begin(in, ss);
     while( step() )
	  ;
     end();
}

template<typename AttributeType>
void TinyParser<AttributeType>::rewrite()
{
   Symbol<AttributeType> sym;
   int id;

   if (isTracing() && traceSymbol(current.id))
      debugString(" : ");

   popL(preContextLength);
   popR(postContextLength);

   int i = preContextLength;
   while(*tablePointer)
   {
      if (isTracing() && traceSymbol(current.id))
      {    debugString(getSymbolName(*tablePointer));
           debugString(" ");
      }

      if (isVariable(*tablePointer))
	   {
	      id = makeId(*tablePointer);
	      sym.attr = lhsAttr(id);
	      sym.id = lhsToken(id);
	      pushL(sym);
	   }
      else
      {
         sym.id = *tablePointer;
         sym.attr = i > 0 ? lhsAttr(-i) : current.attr;
         pushL(sym);
      }
	   ++tablePointer;
	   ++i;
   }

   if (isTracing() && traceSymbol(current.id))
      debugString(". ");

   ++tablePointer;

   int act = -1;
   short rrhslen = 0;
   i = postContextLength;

   while(*tablePointer)
   {
	   if (isVariable(*tablePointer))
	   {
	      id = makeId(*tablePointer);
	      sym.id = lhsToken(id);
	      sym.attr = lhsAttr(id);
	      pushR(sym);
	   }
	   else if (!isAction(*tablePointer))
	   {
	      sym.id = *tablePointer;
	      sym.attr = i > 0 ? lhsAttr(i) : current.attr;
	      pushR(sym);
	   }
	   else
	      act = *tablePointer;

      ++tablePointer;
	   --i;
	   ++rrhslen;
   }

   if (act != -1)
	   action(act);

   ++tablePointer;
   if(*tablePointer != (unsigned short)-1)
   {
      ~tape;
      while(*tablePointer)
      {
	      if (isVariable(*tablePointer))
	      {
	         id = makeId(*tablePointer);
	         sym.id = lhsToken(id);
	         sym.attr = lhsAttr(id);
	         pushR(sym);
         }
	      else if (!isAction(*tablePointer))
	      {
	         sym.id = *tablePointer;
	         sym.attr = i > 0 ? lhsAttr(i) : current.attr;
	         pushR(sym);
	      }
	      else
	         act = *tablePointer;

	      ++tablePointer;
	      --i;
	      ++rrhslen;
      }

      ++tablePointer;
      int act = -1;
      while(*tablePointer)
      {
	      if (isVariable(*tablePointer))
	      {
	         id = makeId(*tablePointer);
	         sym.id = lhsToken(id);
	         sym.attr = lhsAttr(id);
	         pushR(sym);
         } else if (!isAction(*tablePointer))
	      {
	         sym.id = *tablePointer;
	         sym.attr = i > 0 ? lhsAttr(i) : current.attr;
	         pushR(sym);
	      }
	      else
	         act = *tablePointer;

	      ++tablePointer;
	      --i;
	      ++rrhslen;
      }
      ~tape;
   }

   if (isTracing() && traceSymbol(current.id))
   {
      const unsigned short * tabPtr = tablePointer;
      while(rrhslen--)
      {
         debugString(getSymbolName(*--tabPtr));
         debugString(" ");
      }

      debugString("\n");
   }
}

template<typename AttributeType>
int TinyParser<AttributeType>::step()
{
     if (emptyR())
	  return 0;

     popR(current);


     if (symbolType(current.id) == TINY_NONTERMINAL_MASK)
     {
	     if (isTracing() && traceSymbol(current.id) && getVerbose() )
	     {
           debugString("[");
           debugInt(itemCountL());
           debugString(", ");
	        debugInt(itemCountR() + 1);
           debugString("] ");
        }

	     if (!rewriteSymbol(current.id))
	     {
	        if (token(-1) == getErrorSymbol())
		        throw tiny_exception("syntax error in error handler");

	        pushL(current);
	        current.id = getErrorSymbol();
	        pushR(current);
	     }
     }
     else if (symbolType(current.id) == TINY_GENERIC_MASK)
     {
        action(getGenericsTable()[makeId(current.id)]);
     }
     else if (symbolType(current.id) == TINY_TERMINAL_MASK)
     {
	     if (consume)
	        getToken(lookAhead);

	     if (lookAhead.id == current.id || current.id == getAnySymbol())
	     {
	        consume = 1;
	        pushL(lookAhead);

	        if (isTracing())
	        {
              if (getTraceTerminalsFlag() && getVerbose() )
	           {
                 debugString("[");
                 debugInt(itemCountL());
                 debugString(", ");
	              debugInt(itemCountR() + 1);
                 debugString("] ");
              }

	           if (getTraceTerminalsFlag())
	           {
                 debugString("accept ");
	              debugString(getSymbolName(current.id));
	              debugString("\n");
              }
	        }
	     } else if (current.id == getPlaybackSymbol())
        {
	        consume = 0;
           ~tape;
        }
	     else
	     {
	        consume = 0;

	        if (token(-1) == getErrorSymbol())
		        throw tiny_exception("syntax error in error handler");

	        pushL(current);

	        current.id = getErrorSymbol();
	        current.attr = lookAhead.attr;
	        pushR(current);
	     }
     }
     else
     {
	     pushR(current);
	     throw tiny_exception("bad symbol class; probably a bug");
     }

   return 1;
}

#endif
