/*
	Copyright (c) 2001 Maurice Gittens <maurce@gittens.nl>

   This software is distributed under the GNU General
   Public License.
	See the file COPYING for more information on usage and
	redistribution of this file and for a DISCLAIMER OF ALL
	WARRANTIES.

	Written by Maurice Gittens <maurice@gittens.nl>.
*/

#include <tiny_lexer.h>
#include <tiny_grammarsymbol.h>
#include <tiny_rewritablesymbol.h>
#include <tiny_genericsymbol.h>
#include <tiny_variable.h>
#include <tiny_action.h>
#include <tiny_util.h>
#include <iostream>
#include <fstream>
#include <stdexcept>
#include <cctype>
#include <cstdio>

using namespace std;

tiny_lexer::tiny_lexer(const char* n,
                       ofstream& ss,
                       ofstream& hs,
                       unsigned bs)
: sourceStream(ss), headerStream(hs), _eof(false)
{
   bufsize = bs;

	State *state = 0;
	ifstream *ifs = 0;
   errorCount = warningCount = startSymbol =
   playbackSymbol = errorSymbol = anySymbol = 0;

	try
	{
	   ifs = new ifstream(n);

   	if (!(*ifs))
	   {
	      delete ifs;
         ifs = 0;
   	   fatalError(string("can't open input file: ") + n);  // throws an exception
	   }

      sstack.push(new State(n,*ifs, bs));
	}
	catch(...)
	{
	   if (ifs)
		   delete ifs;

		if (state)
		   delete state;

	   throw;
	}

	{
	   typedef tiny_grammarSymbol SType;
      SType* sym;

	   symbolTable.insert(make_pair(string("generics"),
	                     sym = new tiny_grammarSymbol(SType::reserverWord,
	                                            "generics",
	                                            RW_GENERICS,
	                                            0)));
   	sym->setUsed();

	   symbolTable.insert(make_pair(string("nonterminals"),
	                     sym = new tiny_grammarSymbol(SType::reserverWord,
	                                            "nonterminals",
	                                            RW_NONTERMINALS,
	                                            0)));
   	sym->setUsed();

   	symbolTable.insert(make_pair(string("terminals"),
	                     sym = new tiny_grammarSymbol(SType::reserverWord,
	                                            "terminals",
	                                            RW_TERMINALS,
	                                            0)));
   	sym->setUsed();

	   symbolTable.insert(make_pair(string("principles"),
	                     sym = new tiny_grammarSymbol(SType::reserverWord,
	                                            "principles",
	                                            RW_PRINCIPLES,
	                                            0)));
     	sym->setUsed();

	   symbolTable.insert(make_pair(string("grammar"),
	                     sym = new tiny_grammarSymbol(SType::reserverWord,
	                                            "grammar",
	                                            RW_GRAMMAR,
	                                            0)));
   	sym->setUsed();

	   symbolTable.insert(make_pair(string("start"),
	                     sym = new tiny_grammarSymbol(SType::reserverWord,
	                                            "start",
	                                            RW_START,
	                                            0)));
   	sym->setUsed();
	
	   symbolTable.insert(make_pair(string("type"),
	                     sym = new tiny_grammarSymbol(SType::reserverWord,
	                                            "type",
	                                            RW_TYPE,
	                                            0)));
   	sym->setUsed();
	
	   symbolTable.insert(make_pair(string("header"),
	                     sym = new tiny_grammarSymbol(SType::reserverWord,
	                                            "header",
	                                            RW_HEADER,
	                                            0)));
   	sym->setUsed();
	
	   symbolTable.insert(make_pair(string("source"),
	                     sym = new tiny_grammarSymbol(SType::reserverWord,
	                                            "source",
	                                            RW_SOURCE,
	                                            0)));
   	sym->setUsed();
	
	   symbolTable.insert(make_pair(string("error"),
	                     sym = new tiny_grammarSymbol(SType::reserverWord,
	                                            "error",
	                                            RW_ERROR,
	                                            0)));
   	sym->setUsed();
   }	
}

tiny_lexer::~tiny_lexer()
{
}

int tiny_lexer::getToken()
{
   while(!sstack.empty())
	{
      State& top = *sstack.top();
      std::istream& str = top.getStream();
      char *buf = top.buf;
      char *ebuf = top.ebuf;
		int ch;
		bool stop = false;

      lptr = buf;
		*lptr = 0;
		for(;!stop;)
		{
   		ch = str.get();

			switch(ch)
			{
			   case ' ' : case '\t' : case '\f' : case '\r' : continue;
            case '\n' : top.incLineNo(); continue;
				case ';' : return SEMIC;
				case '!' : return PLAYBACK;
				case '|' : return BAR;
				case ':' : return COLON;
				case '.' : return DOT;
				case ',' : return COMMA;
				case '(' : return LP;
				case ')' : return RP;
				break;
				case '/' :
				{
				   if (str.peek() == '*')
					{
					   str.get();
					   eatcomment();
					} else if(str.peek() == '/')
					{
					   str.get();
					   eatline();
					}
					else
					{
				      error(string("unexpected input character: '/'"));
					}
            }
				break;
				case '>' :
				{
				   if (str.peek() == '>')
					{
					   str.get();
					   return RECORD;
					}
					else
					{
				      error(string("unexpected input character: '>'"));
					}
            }
				break;
				case '}' : return RC;
				case '{' :
				{
				   if (str.peek() == '%')
					{
					   str.get();
					   return CODEBLOCK;
					}
					else
					   return LC;

            }
				break;
            case '#' :
				{
					str.getline(buf, bufsize, '\n');

					cerr << "preprocessor directive: " << buf << endl;
				}
            break;
            case EOF :
				   if (!popInputStream())
                  return _eof = true, 0;
					else
					   stop = true;
            break;
			   default :
   		      if (isalpha(ch) || ch == '_')
	   		   {
					   while(isalnum(ch) || ch == '_')
						{
                     *lptr++ = ch;

                     if (lptr >= ebuf) { fatalError("lexeme to long"); }

							ch = str.get();
                  }
						*lptr = 0;
					   str.unget();

						return getIdToken(buf);
   		   	} else
            	{
					   char c[2];
						c[0] = ch;
						c[1] = 0;
				      error(string("unexpected input character: '") + c + '\'');
					}
		   } // end switch
		} // end for
	}

	return 0;
}


int tiny_lexer::getIdToken(const char* str)
{
   map<string, tiny_grammarSymbol*>::const_iterator it =
      symbolTable.find(string("$")+str);

   if (it == symbolTable.end())
   {
      map<string, tiny_grammarSymbol*>::const_iterator it = symbolTable.find(str);

      if (it == symbolTable.end())
      {
         return ID;
      }
	   else
      {
         typedef tiny_grammarSymbol GS;
	
         switch(it->second->getType())
         {
            case GS::reserverWord : return it->second->getId();
            case GS::terminal : return TERMINAL;
            case GS::nonterminal : return NONTERMINAL;
            case GS::generic : return GENERIC;
            case GS::variable : return VARIABLE;
            case GS::action : return ACTION;
            default:
               fatalError("internal error; undefined symbol class");
               return 0;
         }
      }
   }
   else
      return VARIABLE;
}

const char* tiny_lexer::getTokenClass(int token) const
{
   switch(token)
	{
      case RW_NONTERMINALS : return "reservedword nonterminals";
      case RW_TERMINALS : return "reservedword terminals";
      case RW_PRINCIPLES : return "reservedword principles";
      case RW_GRAMMAR : return "reservedword grammar";
      case RW_START : return "reservedword start";
      case RW_TYPE : return "reservedword type";
      case RW_PRODUCTIONS : return "reservedword productions";
      case RW_HEADER : return "reservedword header";
      case RW_SOURCE : return "reservedword source";
      case RW_ERROR : return "reservedword error";
      case CODEBLOCK : return "a block of code";
      case ACTION : return "an action";
      case VARIABLE : return "a variable";
      case TERMINAL : return "a terminal";
      case GENERIC : return "a generic nonterminal";
      case NONTERMINAL : return "a nonterminal";
	   case ID : return "identifier";
	   case SEMIC : return ";";
	   case RECORD : return ">>";
	   case COLON : return ":";
	   case DOT : return ".";
	   case COMMA : return ",";
	   case BAR : return "|";
	   case LC : return "{";
	   case RC : return "}";
	   case LP : return "(";
	   case RP : return ")";
		default : return "<unknown>";
	}
}

bool tiny_lexer::popInputStream()
{
   State* top = sstack.top();
	delete &top->stream;

	sstack.pop();

	delete top;
	return !sstack.empty();
}

void tiny_lexer::setTypeName(const char*symbol)
{
   if (typeName.empty())
   {
      typeName = symbol;
   }
   else
   {
      error("type symbol allready set for this grammar");
   }
}

string tiny_lexer::getCodeBlock()
{
   State& top = *sstack.top();
   istream& str = top.getStream();
   string rv;

	int ch;
	while(EOF != (ch = str.get()))
	{
	   if (ch == '\n')
         top.incLineNo();	
	
	   if (ch == '%' && str.peek() == '}')
		{
         str.get();
			return rv;
		}

		//rv.push_back((char)ch);
		rv += (char)ch;
	}

	if (EOF == ch)
	{
	   error("end of file in codeblock");
	}
	return rv;
}

void tiny_lexer::setAnySymbol()
{
   tiny_grammarSymbol* t = new
      tiny_grammarSymbol(tiny_grammarSymbol::terminal,
	                      "__any__",
	                      getTerminalCount(),
							    getLineNo());

   t->setUsed();
   lterminals.push_back(t);
   anySymbol = t->getEncodedId();

   symbolTable.insert(make_pair(string(getLexeme()), t));
}

void tiny_lexer::setPlaybackSymbol()
{
   tiny_grammarSymbol* t = new
      tiny_grammarSymbol(tiny_grammarSymbol::terminal,
	                      "_playback_",
	                      getTerminalCount(),
							    getLineNo());

   t->setUsed();
   lterminals.push_back(t);
   playbackSymbol = t->getEncodedId();
   symbolTable.insert(make_pair(string("!"), t));
}

tiny_grammarSymbol* tiny_lexer::newTerminal()
{
   return newTerminal(getLexeme());
}

tiny_grammarSymbol* tiny_lexer::newTerminal(const char* name)
{
   tiny_grammarSymbol* t = new
      tiny_grammarSymbol(tiny_grammarSymbol::terminal,
	                      name,
	                      getTerminalCount(),
							    getLineNo());

   if (global)
   {
      gterminals.push_back(t);
   }
   else
   {
      lterminals.push_back(t);
   }

   symbolTable.insert(make_pair(string(name), t));

   return t;
}

tiny_rewritableSymbol* tiny_lexer::newGeneric()
{
   return newGeneric(getLexeme());
}

tiny_rewritableSymbol* tiny_lexer::newGeneric(const char* name)
{
   tiny_genericSymbol* t = new
      tiny_genericSymbol(tiny_genericSymbol::generic,
	                      name,
	                      nonterminals.size(),
							    getLineNo());

   nonterminals.push_back(t);

   t->setAction(newAction(string(" { ") + name + "_nt_();  }\n"));

   symbolTable.insert(make_pair(string(name), t));

   return t;
}

void tiny_lexer::newNonTerminal()
{
   string lexeme = getLexeme();
   tiny_rewritableSymbol* nt = new
      tiny_rewritableSymbol(tiny_grammarSymbol::nonterminal,
	                   lexeme,
	                   nonterminals.size(),
							 getLineNo());

   nonterminals.push_back(nt);
   symbolTable.insert(make_pair(lexeme, nt));
}

void tiny_lexer::clearVariables()
{
   for(vector<string>::const_iterator it = variableNames.begin();
       it != variableNames.end();
       ++it)
   {
      symbolTable.erase(*it);
   }
}

tiny_variable& tiny_lexer::newVariable(const char* name)
{
   string varname = string("$")+name;

   tiny_variable* var = new
      tiny_variable(varname,
	               variables.size(),
						getLineNo());

   variables.push_back(var);
   symbolTable.insert(make_pair(string(varname), var));
   variableNames.push_back(varname);

   return *var;
}

tiny_action& tiny_lexer::newAction(const string& action)
{
   char buf[32];

   sprintf(buf,"#%d", getActionCount());

   tiny_action* act = new
      tiny_action(buf,
	               actions.size(),
						getLineNo());

   actions.push_back(act);
   symbolTable.insert(make_pair(string(buf), act));

   act->setString(action);

   return *act;
}

const tiny_grammarSymbol* tiny_lexer::getSymbol(const char* s) const
{
   SymbolTableType::const_iterator it = symbolTable.find(s);

   if (it != symbolTable.end())
      return it->second;
   else
      return 0;
}

tiny_grammarSymbol* tiny_lexer::getSymbol(const char* s)
{
   SymbolTableType::const_iterator it = symbolTable.find(s);

   if (it != symbolTable.end())
      return it->second;
   else
      return 0;
}

//
//
//
void tiny_lexer::eatcomment()
{
   if (sstack.size())
   {
      State& top = *sstack.top();
      istream& str = top.getStream();

   	int ch;
	   while(EOF != (ch = str.get()))
   	{
	      if (ch == '\n')
            top.incLineNo();	
	
   	   if (ch == '*' && str.peek() == '/')
	   	{
            str.get();
   			return;
	   	}
	   }
   	if (EOF == ch)
	   {
	      error("end of file in comment");
   	}
	}

}

void tiny_lexer::eatline()
{
   if (sstack.size())
   {
      State& top = *sstack.top();
      istream& str = top.getStream();

   	int ch;
	   while(EOF != (ch = str.get()))
   	{
	      if (ch == '\n')
	      {
            top.incLineNo();
            break;
	      }
   	}
   }
}

void tiny_lexer::fatalError(const string& str)
{
   cerr << "fatalError: " << str << endl;
	throw runtime_error(str);
}

void tiny_lexer::error(const string& str, bool eol)
{
   if (sstack.size())
   {
      State& top = *sstack.top();
      cerr << "error: " << top.getStreamName() << ": " << top.getLineNo()
           << " : " << str ;
   } else
   {
      cerr << "error: " << str;
   }

   ++errorCount;
   if (eol)
      cerr << endl;
}

void tiny_lexer::warning(const string& str, bool eol)
{
   if (sstack.size())
   {
      State& top = *sstack.top();
      cerr << "warning: " << top.getStreamName() << ": " << top.getLineNo()
           << " : " << str;
   } else
   {
      cerr << "warning: " << str;
   }

   ++warningCount;

   if (eol)
      cerr << endl;
}


