/*
	Copyright (c) 2001 Maurice Gittens <maurce@gittens.nl>

   This software is distributed under the GNU General
   Public License.
	See the file COPYING for more information on usage and
	redistribution of this file and for a DISCLAIMER OF ALL
	WARRANTIES.

	Written by Maurice Gittens <maurice@gittens.nl>.
*/

#ifndef TINY_LEXER_SEEN
#define TINY_LEXER_SEEN

#include <stack>
#include <fstream>
#include <string>
#include <vector>
#include <map>

#include <tiny_grammarsymbol.h>
class tiny_action;
class tiny_rewritableSymbol;
class tiny_variable;

/**
 *
 */
class tiny_lexer
{
public:
	tiny_lexer(const char* i,
	           std::ofstream& hs,
	           std::ofstream& ss,
	           unsigned = 1024);
   ~tiny_lexer();

	enum Token
	{
      RW_NONTERMINALS = 1, RW_TERMINALS, RW_PRINCIPLES, RW_GENERICS,
      RW_GRAMMAR, RW_START, RW_TYPE, RW_PRODUCTIONS,
      RW_HEADER, RW_SOURCE,
      RW_ERROR, SEMIC, COLON,
      LC, RC, DOT, BAR, RECORD, LP, RP, COMMA,
      CODEBLOCK,
      PLAYBACK,
      ID, ACTION, TERMINAL, VARIABLE, GENERIC, NONTERMINAL,
	};

	struct State
	{
	   State(const char* n, std::ifstream& s, unsigned bs)
		: stream(s), name(n)
		{
		   lineno = 1;
			buf = new char[bs];
			ebuf = buf + bs;
		}

		~State() { delete [] buf;  }

		const char* getLexeme() const { return buf; }
		const char* getStreamName() const { return name.c_str(); }
		std::istream& getStream()  { return stream; }
		unsigned getLineNo() const { return lineno; }

		void incLineNo() { ++lineno; }

	   std::ifstream& stream;
		std::string name;
      unsigned lineno;
      char *buf;
      char *lptr;
      char *ebuf;

	   private:
		   State();
		   State(const State&);
		   State& operator=(const State&);
	};

	void pushInputStream(const char*, std::istream&);
	bool popInputStream();

	int getToken();
	const char* getTokenClass(int) const;
	const char* getLexeme() const { return sstack.top()->getLexeme(); }
	unsigned getLexemeLength() const { return lptr-sstack.top()->buf; }
	unsigned getLineNo() const { return sstack.top()->getLineNo(); }

	void fatalError(const std::string&);
	void error(const std::string&, bool eol = true);
	void warning(const std::string&, bool eol = true);

   const std::string& getHeaderCodeBlock() const { return headerCodeBlock; }
   const std::string& getSourceCodeBlock() const { return sourceCodeBlock; }
   const std::string& getInheritBlock() const { return inheritBlock; }
   void setHeaderCodeBlock(const std::string& s) { headerCodeBlock = s; }
   void setSourceCodeBlock(const std::string& s) { sourceCodeBlock = s; }
   void setInheritBlock(const std::string& ib) { inheritBlock = ib; }
	
	const std::string& getTypeName() const { return typeName; }
	unsigned getStartSymbol() const { return startSymbol; }
	unsigned getErrorSymbol() const { return errorSymbol; }
   unsigned getAnySymbol() const { return anySymbol; }
   unsigned getPlaybackSymbol() const { return playbackSymbol; }

	void setTypeName(const char*);
	void setStartSymbol(unsigned s) { startSymbol = s; }
	void setErrorSymbol(unsigned s) { errorSymbol = s; }
   void setAnySymbol();
   void setPlaybackSymbol();

   unsigned getInitialGlobalTerminalCount() const { return initialGlobalTerminalCount; }
   unsigned getErrorCount() const { return errorCount; }
   unsigned getWarningCount() const { return warningCount; }

	unsigned getLocalTerminalCount() const { return lterminals.size(); }
	unsigned getGlobalTerminalCount() const { return gterminals.size(); }
   unsigned getTerminalCount() const { return lterminals.size() + gterminals.size(); }

   unsigned getVariableCount() const { return variables.size(); }
   unsigned getNonTerminalCount() const { return nonterminals.size(); }
   unsigned getActionCount() const { return actions.size(); }

   const char* getGrammarName() const { return grammarname.c_str(); }
   void setGrammarName(const char* gn) { grammarname = gn; }

   tiny_action& newAction(const std::string&);
   tiny_grammarSymbol* newTerminal();
   tiny_grammarSymbol* newTerminal(const char*);
   tiny_variable& newVariable(const char*);
   tiny_rewritableSymbol* newGeneric();
   tiny_rewritableSymbol* newGeneric(const char*);
   void newNonTerminal();

   void clearVariables();

   tiny_grammarSymbol& getTerminal(unsigned i)
   {
      if (i < gterminals.size())
         return *gterminals[i];
      else
         return *lterminals[i];
   }

   tiny_rewritableSymbol& getNonTerminal(unsigned i) { return *nonterminals[i]; }
   const tiny_grammarSymbol* getSymbol(const char* s) const;
   tiny_grammarSymbol* getSymbol(const char* s);
   tiny_action& getAction(int i) { return *actions[i]; }

	bool eof() const { return _eof; }
   void setIsGlobal(bool g) { global = g; }
   bool getIsGlobal() const { return global; }

   std::string getCodeBlock();
private:
   tiny_lexer(const tiny_lexer&);
   tiny_lexer& operator=(const tiny_lexer&);

   void eatcomment();
   void eatline();

   std::stack<State*> sstack;
   unsigned bufsize;
	std::ofstream& sourceStream;
	std::ofstream& headerStream;

   char *lptr;
   bool _eof;

   std::vector<tiny_grammarSymbol*> lterminals;
   std::vector<tiny_grammarSymbol*> gterminals;
   std::vector<tiny_rewritableSymbol*> nonterminals;
   std::vector<tiny_action*> actions;
   std::vector<tiny_variable*> variables;
   std::vector<std::string> variableNames;

   typedef std::map<std::string, tiny_grammarSymbol*> SymbolTableType;

   SymbolTableType symbolTable;

   std::string grammarname;
   std::string typeName;
	unsigned short startSymbol;
	unsigned short errorSymbol;
   unsigned short anySymbol;
   unsigned short playbackSymbol;
   bool global;

   unsigned short initialGlobalTerminalCount;
   unsigned errorCount;
   unsigned warningCount;

   int getIdToken(const char*);

   std::string headerCodeBlock;
   std::string sourceCodeBlock;
   std::string inheritBlock;
};

#endif
